
# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2022-01-25T13:20:06-0700",
 "dirty": false,
 "error": null,
 "full-revisionid": "e19d47640420a1c5a6ae7d708af97ba2b13178e7",
 "version": "0.3.8"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
